"""Alarm control panel module for zhawss."""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any, Final, cast

from zigpy.zcl.clusters.security import IasAce

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import CLUSTER_HANDLER_EVENT
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_IAS_ACE
from zhaws.server.zigbee.cluster.security import (
    ClusterHandlerStateChangedEvent,
    IasAce as IasAceClusterHandler,
)

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(
    PLATFORM_ENTITIES.strict_match, Platform.ALARM_CONTROL_PANEL
)

SUPPORT_ALARM_ARM_HOME: Final[int] = 1
SUPPORT_ALARM_ARM_AWAY: Final[int] = 2
SUPPORT_ALARM_ARM_NIGHT: Final[int] = 4
SUPPORT_ALARM_TRIGGER: Final[int] = 8
SUPPORT_ALARM_ARM_CUSTOM_BYPASS: Final[int] = 16
SUPPORT_ALARM_ARM_VACATION: Final[int] = 32


class AlarmState(StrEnum):
    DISARMED = "disarmed"
    ARMED_HOME = "armed_home"
    ARMED_AWAY = "armed_away"
    ARMED_NIGHT = "armed_night"
    ARMED_VACATION = "armed_vacation"
    ARMED_CUSTOM_BYPASS = "armed_custom_bypass"
    PENDING = "pending"
    ARMING = "arming"
    DISARMING = "disarming"
    TRIGGERED = "triggered"
    UNKNOWN = "unknown"


IAS_ACE_STATE_MAP = {
    IasAce.PanelStatus.Panel_Disarmed: AlarmState.DISARMED,
    IasAce.PanelStatus.Armed_Stay: AlarmState.ARMED_HOME,
    IasAce.PanelStatus.Armed_Night: AlarmState.ARMED_NIGHT,
    IasAce.PanelStatus.Armed_Away: AlarmState.ARMED_AWAY,
    IasAce.PanelStatus.In_Alarm: AlarmState.TRIGGERED,
}


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_ACE)
class ZHAAlarmControlPanel(PlatformEntity):
    """Alarm Control Panel platform entity implementation."""

    PLATFORM = Platform.ALARM_CONTROL_PANEL

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the alarm control panel."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: IasAceClusterHandler = cast(
            IasAceClusterHandler, cluster_handlers[0]
        )
        self._cluster_handler.panel_code = "1234"
        self._cluster_handler.code_required_arm_actions = False
        self._cluster_handler.max_invalid_tries = 3

        """
        # TODO Once config / storage exist populate these values correctly
        self._cluster_handler.panel_code = async_get_zha_config_value(
            cfg_entry, ZHA_ALARM_OPTIONS, CONF_ALARM_MASTER_CODE, "1234"
        )
        self._cluster_handler.code_required_arm_actions = async_get_zha_config_value(
            cfg_entry, ZHA_ALARM_OPTIONS, CONF_ALARM_ARM_REQUIRES_CODE, False
        )
        self._cluster_handler.max_invalid_tries = async_get_zha_config_value(
            cfg_entry, ZHA_ALARM_OPTIONS, CONF_ALARM_FAILED_TRIES, 3
        )
        """
        self._cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    def handle_cluster_handler_state_changed(
        self, event: ClusterHandlerStateChangedEvent
    ) -> None:
        """Handle state changed on cluster."""
        self.maybe_send_state_changed_event()

    @property
    def code_arm_required(self) -> bool:
        """Whether the code is required for arm actions."""
        return self._cluster_handler.code_required_arm_actions

    async def async_alarm_disarm(self, code: str | None = None, **kwargs: Any) -> None:
        """Send disarm command."""
        self._cluster_handler.arm(IasAce.ArmMode.Disarm, code, 0)
        self.maybe_send_state_changed_event()

    async def async_alarm_arm_home(
        self, code: str | None = None, **kwargs: Any
    ) -> None:
        """Send arm home command."""
        self._cluster_handler.arm(IasAce.ArmMode.Arm_Day_Home_Only, code, 0)
        self.maybe_send_state_changed_event()

    async def async_alarm_arm_away(
        self, code: str | None = None, **kwargs: Any
    ) -> None:
        """Send arm away command."""
        self._cluster_handler.arm(IasAce.ArmMode.Arm_All_Zones, code, 0)
        self.maybe_send_state_changed_event()

    async def async_alarm_arm_night(
        self, code: str | None = None, **kwargs: Any
    ) -> None:
        """Send arm night command."""
        self._cluster_handler.arm(IasAce.ArmMode.Arm_Night_Sleep_Only, code, 0)
        self.maybe_send_state_changed_event()

    async def async_alarm_trigger(self, code: str | None = None, **kwargs: Any) -> None:
        """Send alarm trigger command."""
        self.maybe_send_state_changed_event()

    @property
    def supported_features(self) -> int:
        """Return the list of supported features."""
        return (
            SUPPORT_ALARM_ARM_HOME
            | SUPPORT_ALARM_ARM_AWAY
            | SUPPORT_ALARM_ARM_NIGHT
            | SUPPORT_ALARM_TRIGGER
        )

    @property
    def state(self) -> str:
        """Return the state of the entity."""
        return IAS_ACE_STATE_MAP.get(
            self._cluster_handler.armed_state, AlarmState.UNKNOWN
        )

    def get_state(self) -> dict:
        """Get the state of the alarm control panel."""
        response = super().get_state()
        response["state"] = self.state
        return response

    def to_json(self) -> dict:
        """Return a JSON representation of the alarm control panel."""
        json = super().to_json()
        json["supported_features"] = self.supported_features
        json[
            "code_required_arm_actions"
        ] = self._cluster_handler.code_required_arm_actions
        json["max_invalid_tries"] = self._cluster_handler.max_invalid_tries
        return json



"""Event models for zhawss.

Events are unprompted messages from the server -> client and they contain only the data that is necessary to handle the event.
"""
from __future__ import annotations

from typing import Annotated, Any, Literal, Optional, Union

from pydantic.fields import Field

from zhaws.client.model.types import (
    BaseDevice,
    BatteryState,
    BooleanState,
    CoverState,
    Device,
    DeviceSignature,
    DeviceTrackerState,
    ElectricalMeasurementState,
    FanState,
    GenericState,
    Group,
    LightState,
    LockState,
    ShadeState,
    SmareEnergyMeteringState,
    SwitchState,
)
from zhaws.model import BaseEvent, BaseModel


class MinimalPlatformEntity(BaseModel):
    """Platform entity model."""

    name: str
    unique_id: str
    platform: str


class MinimalEndpoint(BaseModel):
    """Minimal endpoint model."""

    id: int
    unique_id: str


class MinimalDevice(BaseModel):
    """Minimal device model."""

    ieee: str


class Attribute(BaseModel):
    """Attribute model."""

    id: int
    name: str
    value: Any


class MinimalCluster(BaseModel):
    """Minimal cluster model."""

    id: int
    endpoint_attribute: str
    name: str
    endpoint_id: int


class MinimalClusterHandler(BaseModel):
    """Minimal cluster handler model."""

    unique_id: str
    cluster: MinimalCluster


class MinimalGroup(BaseModel):
    """Minimal group model."""

    id: int


class PlatformEntityStateChangedEvent(BaseEvent):
    """Platform entity event."""

    event_type: Literal["platform_entity_event"] = "platform_entity_event"
    event: Literal["platform_entity_state_changed"] = "platform_entity_state_changed"
    platform_entity: MinimalPlatformEntity
    endpoint: Optional[MinimalEndpoint]
    device: Optional[MinimalDevice]
    group: Optional[MinimalGroup]
    state: Annotated[
        Union[
            DeviceTrackerState,
            CoverState,
            ShadeState,
            FanState,
            LockState,
            BatteryState,
            ElectricalMeasurementState,
            LightState,
            SwitchState,
            SmareEnergyMeteringState,
            GenericState,
            BooleanState,
        ],
        Field(discriminator="class_name"),  # noqa: F821
    ]


class ZCLAttributeUpdatedEvent(BaseEvent):
    """ZCL attribute updated event."""

    event_type: Literal["raw_zcl_event"] = "raw_zcl_event"
    event: Literal["attribute_updated"] = "attribute_updated"
    device: MinimalDevice
    cluster_handler: MinimalClusterHandler
    attribute: Attribute
    endpoint: MinimalEndpoint


class ControllerEvent(BaseEvent):
    """Controller event."""

    event_type: Literal["controller_event"] = "controller_event"


class DevicePairingEvent(ControllerEvent):
    """Device pairing event."""

    pairing_status: str


class DeviceJoinedEvent(DevicePairingEvent):
    """Device joined event."""

    event: Literal["device_joined"] = "device_joined"
    ieee: str
    nwk: str


class RawDeviceInitializedEvent(DevicePairingEvent):
    """Raw device initialized event."""

    event: Literal["raw_device_initialized"] = "raw_device_initialized"
    ieee: str
    nwk: str
    manufacturer: str
    model: str
    signature: DeviceSignature


class DeviceFullyInitializedEvent(DevicePairingEvent):
    """Device fully initialized event."""

    event: Literal["device_fully_initialized"] = "device_fully_initialized"
    device: Device
    new_join: bool


class DeviceConfiguredEvent(DevicePairingEvent):
    """Device configured event."""

    event: Literal["device_configured"] = "device_configured"
    device: BaseDevice


class DeviceLeftEvent(ControllerEvent):
    """Device left event."""

    event: Literal["device_left"] = "device_left"
    ieee: str
    nwk: str


class DeviceRemovedEvent(ControllerEvent):
    """Device removed event."""

    event: Literal["device_removed"] = "device_removed"
    device: Device


class DeviceOfflineEvent(BaseEvent):
    """Device offline event."""

    event: Literal["device_offline"] = "device_offline"
    event_type: Literal["device_event"] = "device_event"
    device: MinimalDevice


class DeviceOnlineEvent(BaseEvent):
    """Device online event."""

    event: Literal["device_online"] = "device_online"
    event_type: Literal["device_event"] = "device_event"
    device: MinimalDevice


class ZHAEvent(BaseEvent):
    """ZHA event."""

    event: Literal["zha_event"] = "zha_event"
    event_type: Literal["device_event"] = "device_event"
    device: MinimalDevice
    cluster_handler: MinimalClusterHandler
    endpoint: MinimalEndpoint
    command: str
    args: Union[list, dict]


class GroupRemovedEvent(ControllerEvent):
    """Group removed event."""

    event: Literal["group_removed"] = "group_removed"
    group: Group


class GroupAddedEvent(ControllerEvent):
    """Group added event."""

    event: Literal["group_added"] = "group_added"
    group: Group


class GroupMemberAddedEvent(ControllerEvent):
    """Group member added event."""

    event: Literal["group_member_added"] = "group_member_added"
    group: Group


class GroupMemberRemovedEvent(ControllerEvent):
    """Group member removed event."""

    event: Literal["group_member_removed"] = "group_member_removed"
    group: Group


Events = Annotated[
    Union[
        PlatformEntityStateChangedEvent,
        ZCLAttributeUpdatedEvent,
        DeviceJoinedEvent,
        RawDeviceInitializedEvent,
        DeviceFullyInitializedEvent,
        DeviceConfiguredEvent,
        DeviceLeftEvent,
        DeviceRemovedEvent,
        GroupRemovedEvent,
        GroupAddedEvent,
        GroupMemberAddedEvent,
        GroupMemberRemovedEvent,
        DeviceOfflineEvent,
        DeviceOnlineEvent,
        ZHAEvent,
    ],
    Field(discriminator="event"),  # noqa: F821
]



"""Group for zhaws."""
from __future__ import annotations

import asyncio
import collections
import logging
from typing import TYPE_CHECKING, Any, Callable

import zigpy.exceptions

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.model import STATE_CHANGED, EntityStateChangedEvent
from zhaws.server.util import LogMixin

if TYPE_CHECKING:
    from zigpy.group import Endpoint as ZigpyEndpoint, Group as ZigpyGroup
    from zigpy.types.named import EUI64

    from zhaws.server.platforms import GroupEntity
    from zhaws.server.websocket.server import Server
    from zhaws.server.zigbee.device import Device

_LOGGER = logging.getLogger(__name__)


GroupMemberReference = collections.namedtuple(
    "GroupMemberReference", "ieee endpoint_id"
)


class GroupMember(LogMixin):
    """Composite object that represents a device endpoint in a Zigbee group."""

    def __init__(self, group: Group, device: Device, endpoint_id: int) -> None:
        """Initialize the group member."""
        self._group: Group = group
        self._device: Device = device
        self._endpoint_id: int = endpoint_id

    @property
    def group(self) -> Group:
        """Return the group this member belongs to."""
        return self._group

    @property
    def endpoint_id(self) -> int:
        """Return the endpoint id for this group member."""
        return self._endpoint_id

    @property
    def endpoint(self) -> ZigpyEndpoint:
        """Return the endpoint for this group member."""
        return self._device.device.endpoints.get(self.endpoint_id)

    @property
    def device(self) -> Device:
        """Return the zha device for this group member."""
        return self._device

    @property
    def associated_entities(self) -> list[PlatformEntity]:
        """Return the list of entities that were derived from this endpoint."""
        return [
            platform_entity
            for platform_entity in self._device.platform_entities.values()
            if platform_entity.endpoint.id == self.endpoint_id
        ]

    def to_json(self) -> dict[str, Any]:
        """Get group info."""
        member_info: dict[str, Any] = {}
        member_info["endpoint_id"] = self.endpoint_id
        member_info["device"] = self.device.zha_device_info
        member_info["entities"] = {
            entity.unique_id: entity.to_json() for entity in self.associated_entities
        }
        return member_info

    async def async_remove_from_group(self) -> None:
        """Remove the device endpoint from the provided zigbee group."""
        try:
            await self._device.device.endpoints[self._endpoint_id].remove_from_group(
                self._group.group_id
            )
        except (zigpy.exceptions.ZigbeeException, asyncio.TimeoutError) as ex:
            self.debug(
                "Failed to remove endpoint: %s for device '%s' from group: 0x%04x ex: %s",
                self._endpoint_id,
                self._device.ieee,
                self._group.group_id,
                str(ex),
            )

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"[%s](%s): {msg}"
        args = (f"0x{self._group.group_id:04x}", self.endpoint_id) + args
        _LOGGER.log(level, msg, *args, **kwargs)


class Group(LogMixin):
    """Representation of a Zigbee group."""

    def __init__(self, group: ZigpyGroup, server: Server) -> None:
        """Initialize the group."""
        self._group: ZigpyGroup = group
        self._server: Server = server
        self._group_entities: dict[str, GroupEntity] = {}
        self._entity_unsubs: dict[str, Callable] = {}

    @property
    def group_id(self) -> int:
        """Return the group id."""
        return self._group.group_id

    @property
    def name(self) -> str:
        """Return the name of the group."""
        return self._group.name

    @property
    def group_entities(self) -> dict[str, GroupEntity]:
        """Return the platform entities of the group."""
        return self._group_entities

    @property
    def zigpy_group(self) -> ZigpyGroup:
        """Return the zigpy group."""
        return self._group

    @property
    def members(self) -> list[GroupMember]:
        """Return the ZHA devices that are members of this group."""
        devices: dict[EUI64, Device] = self._server.controller.devices
        return [
            GroupMember(self, devices[member_ieee], endpoint_id)
            for (member_ieee, endpoint_id) in self._group.members.keys()
            if member_ieee in devices
        ]

    def register_group_entity(self, group_entity: GroupEntity) -> None:
        """Register a group entity."""
        if group_entity.unique_id not in self._group_entities:
            self._group_entities[group_entity.unique_id] = group_entity
            self._entity_unsubs[group_entity.unique_id] = group_entity.on_event(
                STATE_CHANGED,
                self._maybe_update_group_members,
            )
        self.update_entity_subscriptions()

    def send_event(self, event: dict[str, Any]) -> None:
        """Send an event from this group."""
        self._server.client_manager.broadcast(event)

    async def _maybe_update_group_members(self, event: EntityStateChangedEvent) -> None:
        """Update the state of the entities that make up the group if they are marked as should poll."""
        tasks = []
        platform_entities = self.get_platform_entities(event.platform)
        for platform_entity in platform_entities:
            if platform_entity.should_poll:
                tasks.append(platform_entity.async_update())
        if tasks:
            await asyncio.gather(*tasks)

    def update_entity_subscriptions(self) -> None:
        """Update the entity event subscriptions.

        Loop over all the entities in the group and update the event subscriptions. Get all of the unique ids
        for both the group entities and the entities that they are compositions of. As we loop through the member
        entities we establish subscriptions for their events if they do not exist. We also add the entity unique id
        to a list for future processing. Once we have processed all group entities we combine the list of unique ids
        for group entities and the platrom entities that we processed. Then we loop over all of the unsub ids and we
        execute the unsubscribe method for each one that isn't in the combined list.
        """
        group_entity_ids = list(self._group_entities.keys())
        processed_platform_entity_ids = []
        for group_entity in self._group_entities.values():
            for platform_entity in self.get_platform_entities(group_entity.PLATFORM):
                processed_platform_entity_ids.append(platform_entity.unique_id)
                if platform_entity.unique_id not in self._entity_unsubs:
                    self._entity_unsubs[
                        platform_entity.unique_id
                    ] = platform_entity.on_event(
                        STATE_CHANGED,
                        group_entity.update,
                    )
        all_ids = group_entity_ids + processed_platform_entity_ids
        existing_unsub_ids = self._entity_unsubs.keys()
        for unsub_id in existing_unsub_ids:
            if unsub_id not in all_ids:
                self._entity_unsubs[unsub_id]()

    async def async_add_members(self, members: list[GroupMemberReference]) -> None:
        """Add members to this group."""
        # TODO handle entity change unsubs and sub for new events
        devices: dict[EUI64, Device] = self._server.controller.devices
        if len(members) > 1:
            tasks = []
            for member in members:
                tasks.append(
                    devices[member.ieee].async_add_endpoint_to_group(
                        member.endpoint_id, self.group_id
                    )
                )
            await asyncio.gather(*tasks)
        else:
            member = members[0]
            await devices[member.ieee].async_add_endpoint_to_group(
                member.endpoint_id, self.group_id
            )
        self.update_entity_subscriptions()

    async def async_remove_members(self, members: list[GroupMemberReference]) -> None:
        """Remove members from this group."""
        # TODO handle entity change unsubs and sub for new events
        devices: dict[EUI64, Device] = self._server.controller.devices
        if len(members) > 1:
            tasks = []
            for member in members:
                tasks.append(
                    devices[member.ieee].async_remove_endpoint_from_group(
                        member.endpoint_id, self.group_id
                    )
                )
            await asyncio.gather(*tasks)
        else:
            member = members[0]
            await devices[member.ieee].async_remove_endpoint_from_group(
                member.endpoint_id, self.group_id
            )
        self.update_entity_subscriptions()

    @property
    def member_entity_ids(self) -> list[PlatformEntity]:
        """Return the platform entities for the members of this group."""
        all_entities: list[PlatformEntity] = []
        for member in self.members:
            entities = member.associated_entities
            for entity in entities:
                all_entities.append(entity)
        return all_entities

    def get_platform_entities(self, platform: str) -> list[PlatformEntity]:
        """Return entities belonging to the specified platform for this group."""
        platform_entities: list[PlatformEntity] = []
        for member in self.members:
            if member.device.is_coordinator:
                continue
            for entity in member.associated_entities:
                if entity.PLATFORM == platform:
                    platform_entities.append(entity)

        return platform_entities

    def to_json(self) -> dict[str, Any]:
        """Get ZHA group info."""
        group_info: dict[str, Any] = {}
        group_info["id"] = self.group_id
        group_info["name"] = self.name
        group_info["members"] = {
            str(member.device.ieee): member.to_json() for member in self.members
        }
        group_info["entities"] = {
            unique_id: entity.to_json()
            for unique_id, entity in self._group_entities.items()
        }
        return group_info

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"[%s](%s): {msg}"
        args = (self.name, self.group_id) + args
        _LOGGER.log(level, msg, *args, **kwargs)



"""WS api for the number platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

ATTR_VALUE = "value"
COMMAND_SET_VALUE = "number_set_value"


@decorators.websocket_command(
    platform_entity_command_schema(
        COMMAND_SET_VALUE,
        {
            vol.Required(ATTR_VALUE): vol.Coerce(float),
        },
    )
)
@decorators.async_response
async def set_value(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Select an option."""
    await execute_platform_entity_command(server, client, message, "async_set_value")


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, set_value)



"""Type information for the websocket api module."""
from __future__ import annotations

from typing import Any, Callable, Coroutine

AsyncWebSocketCommandHandler = Callable[
    [Any, Any, dict[str, Any]], Coroutine[Any, Any, None]
]
WebSocketCommandHandler = Callable[[Any, Any, dict[str, Any]], None]



"""Models that represent types for the zhaws.client.

Types are representations of the objects that exist in zhawss.
"""
from __future__ import annotations

from typing import Annotated, Any, Literal, Optional, Union

from pydantic.fields import Field

from zhaws.event import EventBase
from zhaws.model import BaseModel


class BaseEventedModel(EventBase, BaseModel):
    """Base evented model."""


class Cluster(BaseModel):
    """Cluster model."""

    id: int
    endpoint_attribute: str
    name: str
    endpoint_id: int
    type: str
    commands: list[str]


class ClusterHandler(BaseModel):
    """Cluster handler model."""

    unique_id: str
    cluster: Cluster
    class_name: str
    generic_id: str
    endpoint_id: int
    id: str
    status: str


class Endpoint(BaseModel):
    """Endpoint model."""

    id: int
    unique_id: str


class GenericState(BaseModel):
    """Default state model."""

    class_name: Literal[
        "ZHAAlarmControlPanel",
        "Number",
        "DefaultToneSelectEntity",
        "DefaultSirenLevelSelectEntity",
        "DefaultStrobeLevelSelectEntity",
        "DefaultStrobeSelectEntity",
        "AnalogInput",
        "Humidity",
        "SoilMoisture",
        "LeafWetness",
        "Illuminance",
        "Pressure",
        "Temperature",
        "CarbonDioxideConcentration",
        "CarbonMonoxideConcentration",
        "VOCLevel",
        "PPBVOCLevel",
        "FormaldehydeConcentration",
        "ThermostatHVACAction",
        "SinopeHVACAction",
        "RSSISensor",
        "LQISensor",
        "LastSeenSensor",
    ]
    state: Union[str, bool, int, float, None]


class DeviceTrackerState(BaseModel):
    """Device tracker state model."""

    class_name: Literal["DeviceTracker"] = "DeviceTracker"
    connected: bool
    battery_level: Optional[float]


class BooleanState(BaseModel):
    """Boolean value state model."""

    class_name: Literal[
        "Accelerometer",
        "Occupancy",
        "Opening",
        "BinaryInput",
        "Motion",
        "IASZone",
        "Siren",
    ]
    state: bool


class CoverState(BaseModel):
    """Cover state model."""

    class_name: Literal["Cover"] = "Cover"
    current_position: int
    state: Optional[str]
    is_opening: bool
    is_closing: bool
    is_closed: bool


class ShadeState(BaseModel):
    """Cover state model."""

    class_name: Literal["Shade", "KeenVent"]
    current_position: int
    is_closed: bool


class FanState(BaseModel):
    """Fan state model."""

    class_name: Literal["Fan", "FanGroup"]
    preset_mode: str
    percentage: int


class LockState(BaseModel):
    """Lock state model."""

    class_name: Literal["Lock"] = "Lock"
    is_locked: bool


class BatteryState(BaseModel):
    """Battery state model."""

    class_name: Literal["Battery"] = "Battery"
    state: Optional[Union[str, float, int]]
    battery_size: Optional[str]
    battery_quantity: Optional[str]
    battery_voltage: Optional[float]


class ElectricalMeasurementState(BaseModel):
    """Electrical measurement state model."""

    class_name: Literal[
        "ElectricalMeasurement",
        "ElectricalMeasurementApparentPower",
        "ElectricalMeasurementRMSCurrent",
        "ElectricalMeasurementRMSVoltage",
    ]
    state: Optional[Union[str, float, int]]
    measurement_type: Optional[str]


class LightState(BaseModel):
    """Light state model."""

    class_name: Literal["Light", "HueLight", "ForceOnLight", "LightGroup"]
    on: bool
    brightness: Optional[int]
    hs_color: Optional[tuple[float, float]]
    color_temp: Optional[int]
    effect: Optional[str]
    off_brightness: Optional[int]


class SwitchState(BaseModel):
    """Switch state model."""

    class_name: Literal["Switch", "SwitchGroup"]
    state: bool


class SmareEnergyMeteringState(BaseModel):
    """Smare energy metering state model."""

    class_name: Literal["SmartEnergyMetering", "SmartEnergySummation"]
    state: Optional[Union[str, float, int]]
    device_type: Optional[str]
    status: Optional[str]


class BaseEntity(BaseEventedModel):
    """Base platform entity model."""

    name: str
    unique_id: str
    platform: str
    class_name: str


class BasePlatformEntity(BaseEntity):
    """Base platform entity model."""

    device_ieee: str
    endpoint_id: int


class LockEntity(BasePlatformEntity):
    """Lock entity model."""

    class_name: Literal["Lock"]
    state: LockState


class DeviceTrackerEntity(BasePlatformEntity):
    """Device tracker entity model."""

    class_name: Literal["DeviceTracker"]
    state: DeviceTrackerState


class CoverEntity(BasePlatformEntity):
    """Cover entity model."""

    class_name: Literal["Cover"]
    state: CoverState


class ShadeEntity(BasePlatformEntity):
    """Shade entity model."""

    class_name: Literal["Shade", "KeenVent"]
    state: ShadeState


class BinarySensorEntity(BasePlatformEntity):
    """Binary sensor model."""

    class_name: Literal[
        "Accelerometer", "Occupancy", "Opening", "BinaryInput", "Motion", "IASZone"
    ]
    sensor_attribute: str
    zone_type: Optional[int]
    state: BooleanState


class BaseSensorEntity(BasePlatformEntity):
    """Sensor model."""

    attribute: Optional[str]
    decimals: int
    divisor: int
    multiplier: Union[int, float]
    unit: Optional[int]


class SensorEntity(BaseSensorEntity):
    """Sensor entity model."""

    class_name: Literal[
        "AnalogInput",
        "Humidity",
        "SoilMoisture",
        "LeafWetness",
        "Illuminance",
        "Pressure",
        "Temperature",
        "CarbonDioxideConcentration",
        "CarbonMonoxideConcentration",
        "VOCLevel",
        "PPBVOCLevel",
        "FormaldehydeConcentration",
        "ThermostatHVACAction",
        "SinopeHVACAction",
        "RSSISensor",
        "LQISensor",
        "LastSeenSensor",
    ]
    state: GenericState


class BatteryEntity(BaseSensorEntity):
    """Battery entity model."""

    class_name: Literal["Battery"]
    state: BatteryState


class ElectricalMeasurementEntity(BaseSensorEntity):
    """Electrical measurement entity model."""

    class_name: Literal[
        "ElectricalMeasurement",
        "ElectricalMeasurementApparentPower",
        "ElectricalMeasurementRMSCurrent",
        "ElectricalMeasurementRMSVoltage",
    ]
    state: ElectricalMeasurementState


class SmareEnergyMeteringEntity(BaseSensorEntity):
    """Smare energy metering entity model."""

    class_name: Literal["SmartEnergyMetering", "SmartEnergySummation"]
    state: SmareEnergyMeteringState


class AlarmControlPanelEntity(BasePlatformEntity):
    """Alarm control panel model."""

    class_name: Literal["ZHAAlarmControlPanel"]
    supported_features: int
    code_required_arm_actions: bool
    max_invalid_tries: int
    state: GenericState


class ButtonEntity(BasePlatformEntity):
    """Button model."""

    class_name: Literal["IdentifyButton"]
    command: str


class FanEntity(BasePlatformEntity):
    """Fan model."""

    class_name: Literal["Fan"]
    preset_modes: list[str]
    supported_features: int
    speed_count: int
    state: FanState


class LightEntity(BasePlatformEntity):
    """Light model."""

    class_name: Literal["Light", "HueLight", "ForceOnLight"]
    supported_features: int
    min_mireds: int
    max_mireds: int
    effect_list: Optional[list[str]]
    state: LightState


class NumberEntity(BasePlatformEntity):
    """Number entity model."""

    class_name: Literal["Number"]
    engineer_units: int
    application_type: int
    step: float
    min_value: float
    max_value: float
    name: str
    state: GenericState


class SelectEntity(BasePlatformEntity):
    """Select entity model."""

    class_name: Literal[
        "DefaultToneSelectEntity",
        "DefaultSirenLevelSelectEntity",
        "DefaultStrobeLevelSelectEntity",
        "DefaultStrobeSelectEntity",
    ]
    enum: str
    options: list[str]
    state: GenericState


class SirenEntity(BasePlatformEntity):
    """Siren entity model."""

    class_name: Literal["Siren"]
    available_tones: Optional[Union[list[Union[int, str]], dict[int, str]]]
    supported_features: int
    state: BooleanState


class SwitchEntity(BasePlatformEntity):
    """Switch entity model."""

    class_name: Literal["Switch"]
    state: SwitchState


class DeviceSignatureEndpoint(BaseModel):
    """Device signature endpoint model."""

    profile_id: Optional[str]
    device_type: Optional[str]
    input_clusters: list[str]
    output_clusters: list[str]


class NodeDescriptor(BaseModel):
    """Node descriptor model."""

    logical_type: int
    complex_descriptor_available: bool
    user_descriptor_available: bool
    reserved: int
    aps_flags: int
    frequency_band: int
    mac_capability_flags: int
    manufacturer_code: int
    maximum_buffer_size: int
    maximum_incoming_transfer_size: int
    server_mask: int
    maximum_outgoing_transfer_size: int
    descriptor_capability_field: int


class DeviceSignature(BaseModel):
    node_descriptor: Optional[NodeDescriptor]
    manufacturer: Optional[str]
    model: Optional[str]
    endpoints: dict[int, DeviceSignatureEndpoint]


class BaseDevice(BaseModel):
    """Base device model."""

    ieee: str
    nwk: str
    manufacturer: str
    model: str
    name: str
    quirk_applied: bool
    quirk_class: Union[str, None]
    manufacturer_code: int
    power_source: str
    lqi: Union[int, None]
    rssi: Union[int, None]
    last_seen: str
    available: bool
    device_type: Literal["Coordinator", "Router", "EndDevice"]
    signature: DeviceSignature


class Device(BaseDevice):
    """Device model."""

    entities: dict[
        str,
        Annotated[
            Union[
                SirenEntity,
                SelectEntity,
                NumberEntity,
                LightEntity,
                FanEntity,
                ButtonEntity,
                AlarmControlPanelEntity,
                SensorEntity,
                BinarySensorEntity,
                DeviceTrackerEntity,
                ShadeEntity,
                CoverEntity,
                LockEntity,
                SwitchEntity,
                BatteryEntity,
                ElectricalMeasurementEntity,
                SmareEnergyMeteringEntity,
            ],
            Field(discriminator="class_name"),  # noqa: F821
        ],
    ]
    neighbors: list[Any]
    device_automation_triggers: dict[str, dict[str, Any]]


class GroupEntity(BaseEntity):
    """Group entity model."""

    group_id: int
    state: Any


class LightGroupEntity(GroupEntity):
    """Group entity model."""

    class_name: Literal["LightGroup"]
    state: LightState


class FanGroupEntity(GroupEntity):
    """Group entity model."""

    class_name: Literal["FanGroup"]
    state: FanState


class SwitchGroupEntity(GroupEntity):
    """Group entity model."""

    class_name: Literal["SwitchGroup"]
    state: SwitchState


class GroupMember(BaseModel):
    """Group member model."""

    endpoint_id: int
    device: Device
    entities: dict[
        str,
        Annotated[
            Union[
                SirenEntity,
                SelectEntity,
                NumberEntity,
                LightEntity,
                FanEntity,
                ButtonEntity,
                AlarmControlPanelEntity,
                SensorEntity,
                BinarySensorEntity,
                DeviceTrackerEntity,
                ShadeEntity,
                CoverEntity,
                LockEntity,
                SwitchEntity,
                BatteryEntity,
                ElectricalMeasurementEntity,
                SmareEnergyMeteringEntity,
            ],
            Field(discriminator="class_name"),  # noqa: F821
        ],
    ]


class Group(BaseModel):
    """Group model."""

    name: str
    id: int
    members: dict[str, GroupMember]
    entities: dict[
        str,
        Annotated[
            Union[LightGroupEntity, FanGroupEntity, SwitchGroupEntity],
            Field(discriminator="class_name"),  # noqa: F821
        ],
    ]


class GroupMemberReference(BaseModel):
    """Group member reference model."""

    ieee: str
    endpoint_id: int



"""Configuration utilities for the zhaws server."""
from __future__ import annotations

from zhaws.server.config.model import ServerConfiguration


def zigpy_config(config: ServerConfiguration) -> dict:
    """Return a dict representing the zigpy configuration from the zhaws ServerCOnfiguration."""
    return {
        "database_path": config.zigpy_configuration.database_path,
        "device": config.radio_configuration.dict(exclude={"type": True}),
    }



"""Websocket api for zhawss."""
from __future__ import annotations

from typing import TYPE_CHECKING, Union, cast

import voluptuous

from zhaws.server.const import WEBSOCKET_API
from zhaws.server.websocket.api.types import WebSocketCommandHandler

if TYPE_CHECKING:
    from zhaws.server.websocket.server import Server


def register_api_command(
    server: Server,
    command_or_handler: Union[str, WebSocketCommandHandler],
    handler: Union[WebSocketCommandHandler, None] = None,
    schema: Union[voluptuous.Schema, None] = None,
) -> None:
    """Register a websocket command."""
    # pylint: disable=protected-access
    if handler is None:
        handler = cast(WebSocketCommandHandler, command_or_handler)
        command = handler._ws_command  # type: ignore[attr-defined]
        schema = handler._ws_schema  # type: ignore[attr-defined]
    else:
        command = command_or_handler
    if (handlers := server.data.get(WEBSOCKET_API)) is None:
        handlers = server.data[WEBSOCKET_API] = {}
    handlers[command] = (handler, schema)



"""WS API for the light platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.platforms.light import (
    ATTR_BRIGHTNESS,
    ATTR_COLOR_TEMP,
    ATTR_EFFECT,
    ATTR_FLASH,
    ATTR_HS_COLOR,
    ATTR_TRANSITION,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server


class LightAPICommands(StrEnum):
    """Light API commands."""

    LIGHT_TURN_ON = "light_turn_on"
    LIGHT_TURN_OFF = "light_turn_off"


COLOR_GROUP = "Color descriptors"

FLASH_SHORT = "short"
FLASH_LONG = "long"

VALID_TRANSITION = vol.All(vol.Coerce(float), vol.Clamp(min=0, max=6553))
VALID_BRIGHTNESS = vol.All(vol.Coerce(int), vol.Clamp(min=0, max=255))
VALID_FLASH = vol.In([FLASH_SHORT, FLASH_LONG])


@decorators.websocket_command(
    platform_entity_command_schema(
        LightAPICommands.LIGHT_TURN_ON,
        {
            vol.Optional(ATTR_BRIGHTNESS): VALID_BRIGHTNESS,
            vol.Optional(ATTR_TRANSITION): VALID_TRANSITION,
            vol.Optional(ATTR_FLASH): VALID_FLASH,
            vol.Optional(ATTR_EFFECT): str,
            vol.Exclusive(ATTR_HS_COLOR, COLOR_GROUP): vol.All(
                vol.Coerce(tuple),
                vol.ExactSequence(
                    (
                        vol.All(vol.Coerce(float), vol.Range(min=0, max=360)),
                        vol.All(vol.Coerce(float), vol.Range(min=0, max=100)),
                    )
                ),
            ),
            vol.Exclusive(ATTR_COLOR_TEMP, COLOR_GROUP): vol.All(
                vol.Coerce(int), vol.Range(min=1)
            ),
        },
    )
)
@decorators.async_response
async def turn_on(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn on the light."""
    await execute_platform_entity_command(server, client, message, "async_turn_on")


@decorators.websocket_command(
    platform_entity_command_schema(
        LightAPICommands.LIGHT_TURN_OFF,
        {
            vol.Optional(ATTR_TRANSITION): VALID_TRANSITION,
            vol.Optional(ATTR_FLASH): VALID_FLASH,
        },
    )
)
@decorators.async_response
async def turn_off(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn on the light."""
    await execute_platform_entity_command(server, client, message, "async_turn_off")


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, turn_on)
    register_api_command(server, turn_off)



from websockets.__main__ import main as websockets_cli

if __name__ == "__main__":
    # "Importing this module enables command line editing using GNU readline."
    import readline  # noqa: F401

    websockets_cli()



"""WS api for the alarm control panel platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Final

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server


class AlarmControlPanelCommands(StrEnum):
    """Alarm control panel commands."""

    DISARM = "alarm_control_panel_disarm"
    ARM_HOME = "alarm_control_panel_arm_home"
    ARM_AWAY = "alarm_control_panel_arm_away"
    ARM_NIGHT = "alarm_control_panel_arm_night"
    TRIGGER = "alarm_control_panel_trigger"


ATTR_CODE: Final[str] = "code"


@decorators.websocket_command(
    platform_entity_command_schema(
        AlarmControlPanelCommands.DISARM,
        {
            vol.Optional(ATTR_CODE): str,
        },
    )
)
@decorators.async_response
async def disarm(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Disarm the alarm control panel."""
    await execute_platform_entity_command(server, client, message, "async_alarm_disarm")


@decorators.websocket_command(
    platform_entity_command_schema(
        AlarmControlPanelCommands.ARM_HOME,
        {
            vol.Optional(ATTR_CODE): str,
        },
    )
)
@decorators.async_response
async def arm_home(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Arm the alarm control panel in home mode."""
    await execute_platform_entity_command(
        server, client, message, "async_alarm_arm_home"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        AlarmControlPanelCommands.ARM_AWAY,
        {
            vol.Optional(ATTR_CODE): str,
        },
    )
)
@decorators.async_response
async def arm_away(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Arm the alarm control panel in away mode."""
    await execute_platform_entity_command(
        server, client, message, "async_alarm_arm_away"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        AlarmControlPanelCommands.ARM_NIGHT,
        {
            vol.Optional(ATTR_CODE): str,
        },
    )
)
@decorators.async_response
async def arm_night(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Arm the alarm control panel in night mode."""
    await execute_platform_entity_command(
        server, client, message, "async_alarm_arm_night"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        AlarmControlPanelCommands.TRIGGER,
        {
            vol.Optional(ATTR_CODE): str,
        },
    )
)
@decorators.async_response
async def trigger(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Trigger the alarm control panel."""
    await execute_platform_entity_command(
        server, client, message, "async_alarm_trigger"
    )


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, disarm)
    register_api_command(server, arm_home)
    register_api_command(server, arm_away)
    register_api_command(server, arm_night)
    register_api_command(server, trigger)



"""Entity helpers for the zhaws server."""
from __future__ import annotations

from typing import Any, Callable, Iterator


def find_state_attributes(states: list[dict], key: str) -> Iterator[Any]:
    """Find attributes with matching key from states."""
    for state in states:
        if (value := state.get(key)) is not None:
            yield value


def mean_int(*args: Any) -> int:
    """Return the mean of the supplied values."""
    return int(sum(args) / len(args))


def mean_tuple(*args: Any) -> tuple:
    """Return the mean values along the columns of the supplied values."""
    return tuple(sum(x) / len(x) for x in zip(*args))


def reduce_attribute(
    states: list[dict],
    key: str,
    default: Any | None = None,
    reduce: Callable[..., Any] = mean_int,
) -> Any:
    """Find the first attribute matching key from states.

    If none are found, return default.
    """
    attrs = list(find_state_attributes(states, key))

    if not attrs:
        return default

    if len(attrs) == 1:
        return attrs[0]

    return reduce(*attrs)



"""ZHAWSS websocket server."""
from __future__ import annotations

import asyncio
import logging
from types import TracebackType
from typing import TYPE_CHECKING, Any

import voluptuous
import websockets

from zhaws.server.config.model import ServerConfiguration
from zhaws.server.const import COMMAND, APICommands
from zhaws.server.platforms import discovery
from zhaws.server.platforms.api import load_platform_entity_apis
from zhaws.server.platforms.discovery import PLATFORMS
from zhaws.server.websocket.api import decorators, register_api_command
from zhaws.server.websocket.client import ClientManager
from zhaws.server.zigbee.api import load_api as load_zigbee_controller_api
from zhaws.server.zigbee.controller import Controller

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client

_LOGGER = logging.getLogger(__name__)


class Server:
    """ZHAWSS server implementation."""

    def __init__(self, *, configuration: ServerConfiguration) -> None:
        """Initialize the server."""
        self._config = configuration
        self._ws_server: websockets.Serve | None = None
        self._controller: Controller = Controller(self)
        self._client_manager: ClientManager = ClientManager(self)
        self._stopped_event: asyncio.Event = asyncio.Event()
        self._tracked_tasks: list[asyncio.Task] = []
        self.data: dict[Any, Any] = {}
        for platform in PLATFORMS:
            self.data.setdefault(platform, [])
        self._register_api_commands()
        discovery.PROBE.initialize(self)
        discovery.GROUP_PROBE.initialize(self)

    @property
    def is_serving(self) -> bool:
        """Returns whether or not the websocket server is serving."""
        return self._ws_server is not None and self._ws_server.is_serving

    @property
    def controller(self) -> Controller:
        """Return the zigbee application controller."""
        return self._controller

    @property
    def client_manager(self) -> ClientManager:
        """Return the zigbee application controller."""
        return self._client_manager

    @property
    def config(self) -> ServerConfiguration:
        """Return the server configuration."""
        return self._config

    async def start_server(self) -> None:
        """Start the websocket server."""
        assert self._ws_server is None
        self._stopped_event.clear()
        self._ws_server = await websockets.serve(
            self.client_manager.add_client,
            self._config.host,
            self._config.port,
            logger=_LOGGER,
        )
        if self._config.network_auto_start:
            await self._controller.start_network()

    async def wait_closed(self) -> None:
        """Waits until the server is not running."""
        await self._stopped_event.wait()
        _LOGGER.info("Server stopped. Completing remaining tasks...")
        await asyncio.gather(*self._tracked_tasks, return_exceptions=True)

    async def stop_server(self) -> None:
        """Stop the websocket server."""
        if self._ws_server is None:
            self._stopped_event.set()
            return

        assert self._ws_server is not None

        if self._controller.is_running:
            await self._controller.stop_network()

        self._ws_server.close()
        await self._ws_server.wait_closed()
        self._ws_server = None

        self._stopped_event.set()

    async def __aenter__(self) -> Server:
        await self.start_server()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        await self.stop_server()

    def _register_api_commands(self) -> None:
        """Load server API commands."""
        from zhaws.server.websocket.client import load_api as load_client_api

        register_api_command(self, stop_server)
        load_zigbee_controller_api(self)
        load_platform_entity_apis(self)
        load_client_api(self)


@decorators.websocket_command(
    {
        voluptuous.Required(COMMAND): str(APICommands.STOP_SERVER),
    }
)
@decorators.async_response
async def stop_server(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Stop the Zigbee network."""
    client.send_result_success(message)
    await server.stop_server()



"""Proxy object for the client side objects."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from zhaws.client.model.events import PlatformEntityStateChangedEvent
from zhaws.client.model.types import ButtonEntity
from zhaws.event import EventBase

if TYPE_CHECKING:
    from zhaws.client.client import Client
    from zhaws.client.controller import Controller
    from zhaws.client.model.types import Device as DeviceModel, Group as GroupModel


class BaseProxyObject(EventBase):
    """BaseProxyObject for the zhaws.client."""

    def __init__(self, controller: Controller, client: Client):
        """Initialize the BaseProxyObject class."""
        super().__init__()
        self._controller: Controller = controller
        self._client: Client = client
        self._proxied_object: GroupModel | DeviceModel

    @property
    def controller(self) -> Controller:
        """Return the controller."""
        return self._controller

    @property
    def client(self) -> Client:
        """Return the client."""
        return self._client

    def emit_platform_entity_event(
        self, event: PlatformEntityStateChangedEvent
    ) -> None:
        """Proxy the firing of an entity event."""
        entity = self._proxied_object.entities.get(event.platform_entity.unique_id)
        if entity is None:
            raise ValueError("Entity not found: %s", event.platform_entity.unique_id)
        if not isinstance(entity, ButtonEntity):
            entity.state = event.state
        self.emit(f"{event.platform_entity.unique_id}_{event.event}", event)


class GroupProxy(BaseProxyObject):
    """Group proxy for the zhaws.client."""

    def __init__(self, group_model: GroupModel, controller: Controller, client: Client):
        """Initialize the GroupProxy class."""
        super().__init__(controller, client)
        self._proxied_object: GroupModel = group_model

    @property
    def group_model(self) -> GroupModel:
        """Return the group model."""
        return self._proxied_object

    @group_model.setter
    def group_model(self, group_model: GroupModel) -> None:
        """Set the group model."""
        self._proxied_object = group_model

    def __repr__(self) -> str:
        return self._proxied_object.__repr__()


class DeviceProxy(BaseProxyObject):
    """Device proxy for the zhaws.client."""

    def __init__(
        self, device_model: DeviceModel, controller: Controller, client: Client
    ):
        """Initialize the DeviceProxy class."""
        super().__init__(controller, client)
        self._proxied_object: DeviceModel = device_model

    @property
    def device_model(self) -> DeviceModel:
        """Return the device model."""
        return self._proxied_object

    @device_model.setter
    def device_model(self, device_model: DeviceModel) -> None:
        """Set the device model."""
        self._proxied_object = device_model

    @property
    def device_automation_triggers(self) -> dict[tuple[str, str], dict[str, Any]]:
        """Return the device automation triggers."""
        model_triggers = self._proxied_object.device_automation_triggers
        return {
            (key.split("~")[0], key.split("~")[1]): value
            for key, value in model_triggers.items()
        }

    def __repr__(self) -> str:
        return self._proxied_object.__repr__()



"""Setup module for zha-websocket-server"""

import pathlib

from setuptools import find_packages, setup

setup(
    name="zhaws",
    version="2022.02.13",
    description="Library implementing a Zigbee websocket server",
    long_description=(pathlib.Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/zigpy/zha-websocket-server",
    author="David F. Mulcahey",
    author_email="david.mulcahey@icloud.com",
    license="GPL-3.0",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "colorlog",
        "pydantic",
        "websockets",
        "voluptuous",
        "zigpy==0.43.0",
        "bellows==0.29.0",
        "zha-quirks==0.0.66",
        "zigpy-deconz==0.14.0",
        "zigpy-xbee==0.14.0",
        "zigpy-zigate==0.8.0",
        "zigpy-znp==0.7.0",
    ],
    extras_require={"server": ["uvloop"]},
    package_data={"": ["appdb_schemas/schema_v*.sql"]},
)



"""WS API for the fan platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

ATTR_SPEED = "speed"
ATTR_PERCENTAGE = "percentage"
ATTR_PRESET_MODE = "preset_mode"


class FanAPICommands(StrEnum):
    """Light API commands."""

    TURN_ON = "fan_turn_on"
    TURN_OFF = "fan_turn_off"
    SET_PERCENTAGE = "fan_set_percentage"
    SET_PRESET_MODE = "fan_set_preset_mode"


@decorators.websocket_command(
    platform_entity_command_schema(
        FanAPICommands.TURN_ON,
        {
            vol.Optional(ATTR_SPEED): str,
            vol.Optional(ATTR_PERCENTAGE): vol.All(
                vol.Coerce(int), vol.Range(min=0, max=100)
            ),
            vol.Optional(ATTR_PRESET_MODE): str,
        },
    )
)
@decorators.async_response
async def turn_on(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn fan on."""
    await execute_platform_entity_command(server, client, message, "async_turn_on")


@decorators.websocket_command(platform_entity_command_schema(FanAPICommands.TURN_OFF))
@decorators.async_response
async def turn_off(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn fan off."""
    await execute_platform_entity_command(server, client, message, "async_turn_off")


@decorators.websocket_command(
    platform_entity_command_schema(
        FanAPICommands.SET_PERCENTAGE,
        {
            vol.Required(ATTR_PERCENTAGE): vol.All(
                vol.Coerce(int), vol.Range(min=0, max=100)
            ),
        },
    )
)
@decorators.async_response
async def set_percentage(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the fan speed percentage."""
    await execute_platform_entity_command(
        server, client, message, "async_set_percentage"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        FanAPICommands.TURN_OFF,
        {
            vol.Required(ATTR_PRESET_MODE): str,
        },
    )
)
@decorators.async_response
async def set_preset_mode(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the fan preset mode."""
    await execute_platform_entity_command(
        server, client, message, "async_set_preset_mode"
    )


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, turn_on)
    register_api_command(server, turn_off)
    register_api_command(server, set_percentage)
    register_api_command(server, set_preset_mode)
